"""Templates module tests."""
import pytest

from homer.exceptions import HomerError
from homer.templates import Renderer
from homer.tests import get_fixture_path


class TestRenderer:
    """Render class tests."""

    def setup_method(self):
        """Initialize the test instances."""
        # pylint: disable=attribute-defined-outside-init
        self.renderer = Renderer(get_fixture_path('templates'), '')

    def test_render_ok(self):
        """Should return the rendered template."""
        assert self.renderer.render('valid', {'test_key': 'test_value'}) == 'test_value;'

    def test_render_syntax_error(self):
        """Should raise HomerError if the template has a syntax error."""
        with pytest.raises(HomerError, match='Syntax error on template syntax_error.conf'):
            self.renderer.render('syntax_error', {})

    def test_render_parse_error(self):
        """Should raise HomerError if the template cannot be parsed."""
        with pytest.raises(HomerError, match='Could not render template non_existent.conf'):
            self.renderer.render('non_existent', {})
        with pytest.raises(HomerError, match='Could not render template key_error.conf'):
            self.renderer.render('key_error', {})



"""Capirca module."""

import logging

from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Mapping

import pynetbox

from capirca.lib import juniper, junipersrx, naming, policy
from requests.exceptions import RequestException

from homer.exceptions import HomerError

logger = logging.getLogger(__name__)


class CapircaGenerate():
    """Class to generate ACLs with Capirca."""

    def __init__(self, config: Mapping, device_policies: List[str], netbox: pynetbox.api):
        """Initialize the instance.

        Arguments:
            config (dict): the Homer config.
            device_policies (list): List of Capirca policies to generate.
            netbox (pynetbox.api): the Netbox API instance.

        """
        self._config = config
        self._device_policies = device_policies
        self._public_policies_dir = Path(self._config['base_paths']['public'], 'policies')
        self._private_base_path = self._config['base_paths'].get('private', None)
        definitions_path = Path(self._config['base_paths']['public'], 'definitions')
        self.definitions = naming.Naming(str(definitions_path))
        # If we want to use Netbox as network definition source
        if self._config.get('capirca', {}).get('netbox_definitons', True) and netbox:
            try:
                script_result = netbox.extras.scripts.get('capirca.GetHosts').result
                if str(script_result.status) != 'Completed':
                    raise HomerError(f'Netbox capirca.GetHosts script status is: {script_result.status}.')
                runtime = datetime.fromisoformat(script_result.completed[:-1])  # To remove the final Z
                now = datetime.utcnow()
                # Warn the user if the Netbox data is 3 day old or more
                if runtime < now - timedelta(days=3):
                    logger.warning('Netbox capirca.GetHosts script is > 3 days old.')
                netbox_definitons = script_result.data.output
            except (pynetbox.RequestError, RequestException) as e:
                raise HomerError('Make sure homer can reach the capirca.GetHosts script on Netbox.') from e
            # ParseNetworkList expects an array of lines, while Netbox API returns a string with \n
            self.definitions.ParseNetworkList(netbox_definitons.splitlines())

    def generate_acls(self) -> List[str]:  # noqa, mccabe: MC0001 too complex
        """Generate the ACLs using Capirca lib.

        Returns:
            list: a list of policies as strings in the proper format.

        """
        generated_acls = []
        if self._private_base_path:
            private_policies_dir = Path(self._private_base_path, 'policies')

        # Store failures if any
        failures = []
        # Iterate over all policies defined for a given device
        for policy_name in self._device_policies:
            # We don't know yet if the files below exist
            public_policy_file = Path(self._public_policies_dir, policy_name + '.pol')
            private_policy_file = Path()
            if self._private_base_path:
                private_policy_file = Path(private_policies_dir, policy_name + '.pol')
            # If same file in both private and public, prefer private
            if private_policy_file.is_file():
                policy_file = private_policy_file
                policies_dir = private_policies_dir
            elif public_policy_file.is_file():
                policy_file = public_policy_file
                policies_dir = self._public_policies_dir
            else:
                failures.append(f"Can't find Capirca policy file {policy_name}.")
                continue
            policy_text = policy_file.read_text(encoding='utf-8')
            try:
                policy_object = policy.ParsePolicy(policy_text, self.definitions,
                                                   optimize=True,
                                                   shade_check=False,
                                                   base_dir=policies_dir)
            except policy.ShadingError as e:
                # Term "hiding" another term
                failures.append(f"Shading errors for {policy_name}: {e}.")
                continue
            except (policy.Error, naming.Error) as e:
                failures.append(f"Error parsing {policy_name}: {e}.")
                continue

            # https://github.com/google/capirca/blob/master/capirca/aclgen.py#L222
            platforms = set()
            for header in policy_object.headers:
                platforms.update(header.platforms)
            try:
                found_platform = False
                if 'juniper' in platforms:
                    found_platform = True
                    generated_acls.append(str(juniper.Juniper(policy_object, exp_info=2)))
                if 'srx' in platforms:
                    found_platform = True
                    generated_acls.append(str(junipersrx.JuniperSRX(policy_object, exp_info=2)))
                if not found_platform:
                    failures.append(f"Unknown platform: {platforms}.")

            except (juniper.Error) as e:
                failures.append(f"Error generating {policy_name}: {e}.")
                continue
        if failures:
            raise HomerError("Capirca error(s)\n" + '\n'.join(failures))

        return generated_acls



"""CLI module tests."""
import argparse

import yaml

from homer import cli
from homer.config import load_yaml_config
from homer.tests import get_fixture_path


def test_argument_parser():
    """It should return an ArgumentParser parser."""
    assert isinstance(cli.argument_parser(), argparse.ArgumentParser)


def test_main(tmp_path):
    """It should execute the whole program based on CLI arguments."""
    output = tmp_path / 'output'
    output.mkdir()
    config = load_yaml_config(get_fixture_path('cli', 'config.yaml'))
    config['base_paths']['output'] = str(output)
    config_path = tmp_path / 'config.yaml'
    with open(str(config_path), 'w', encoding='utf-8') as f:
        yaml.dump(config, f)

    assert cli.main(['-c', str(config_path), 'device1.example.com', 'generate']) == 0



"""Config module tests."""
import ipaddress

import pytest

from homer.config import HierarchicalConfig, load_yaml_config
from homer.devices import Device
from homer.exceptions import HomerError
from homer.tests import get_fixture_path


@pytest.mark.parametrize('name', ('empty.yaml', 'non-existent.yaml'))
def test_load_yaml_config_no_content(name):
    """Loading an empty or non-existent config should return an empty dictionary."""
    assert {} == load_yaml_config(get_fixture_path('config', name))


def test_load_yaml_config_raise():
    """Loading an invalid config should raise Exception."""
    with pytest.raises(HomerError, match='Could not load config file'):
        load_yaml_config(get_fixture_path('config', 'invalid.yaml'))


def test_load_yaml_config_valid():
    """Loading a valid config should return its content."""
    config = load_yaml_config(get_fixture_path('config', 'valid.yaml'))
    assert 'key' in config


def test_load_yaml_config_ipaddress_objects():
    """It should return the configuration with IP addresses, networks and interfaces converted into objects."""
    config = load_yaml_config(get_fixture_path('config', 'ipaddress.yaml'))
    assert isinstance(config['ipv4'], ipaddress.IPv4Address)
    assert isinstance(config['ipv6'], ipaddress.IPv6Address)
    assert isinstance(config['networkv4'], ipaddress.IPv4Network)
    assert isinstance(config['networkv6'], ipaddress.IPv6Network)
    assert isinstance(config['interfacev4'], ipaddress.IPv4Interface)
    assert isinstance(config['interfacev6'], ipaddress.IPv6Interface)
    assert isinstance(config['non_parsable_interface'], str)
    assert isinstance(config['non_parsable_ip'], str)


def test_hierarchical_config_get_no_private():
    """Calling the get() method on an instance of HierarchicalConfig should return the config for a given Device."""
    device = Device('device1.example.com', {'role': 'roleA', 'site': 'siteA'}, {'device_key': 'device1_value'}, {})
    config = HierarchicalConfig(get_fixture_path('public'))
    expected = {'common_key': 'common_value', 'role_key': 'roleA_value', 'site_key': 'siteA_value',
                'device_key': 'device1_value', 'metadata': {'role': 'roleA', 'site': 'siteA'},
                'hostname': 'device1.example.com'}
    assert config.get(device) == expected


def test_hierarchical_config_get_with_private():
    """Calling the get() method on an instance of HierarchicalConfig should include the private config, if any."""
    device = Device('device1.example.com', {'role': 'roleA', 'site': 'siteA'}, {'device_key': 'device1_value'},
                    {'device_private_key': 'device1_private_value'})
    config = HierarchicalConfig(get_fixture_path('public'), private_base_path=get_fixture_path('private'))
    expected = {'common_key': 'common_value', 'role_key': 'roleA_value', 'site_key': 'siteA_value',
                'device_key': 'device1_value', 'common_private_key': 'common_private_value',
                'role_private_key': 'roleA_private_value', 'site_private_key': 'siteA_private_value',
                'device_private_key': 'device1_private_value', 'metadata': {'role': 'roleA', 'site': 'siteA'},
                'hostname': 'device1.example.com'}
    assert config.get(device) == expected


def test_hierarchical_config_get_duplicate_keys():
    """If there are duplicate keys between public and private configuration."""
    device = Device('device1.example.com', {'role': 'roleA', 'site': 'siteA'}, {'device_key': 'device_value'},
                    {'role_key': 'duplicated_key'})
    config = HierarchicalConfig(get_fixture_path('public'), private_base_path=get_fixture_path('private'))
    with pytest.raises(HomerError, match=r'Configuration key\(s\) found in both public and private config'):
        config.get(device)


def test_uncopiable_object():
    """It should return the configuration also if the device metadata has uncopiable objects."""
    class Uncopiable:
        """Uncopiable class."""

        def __deepcopy__(self, _):
            """Make the class uncopiable."""
            raise TypeError('Uncopiable')

    device = Device('device1.example.com', {'role': 'roleA', 'site': 'siteA', 'uncopiable': Uncopiable()}, {}, {})
    config = HierarchicalConfig(get_fixture_path('public'), private_base_path=get_fixture_path('private'))
    config.get(device)



"""Devices module."""
import fnmatch
import logging

from collections import UserDict
from operator import attrgetter
from typing import List, Mapping, NamedTuple, Optional


Device = NamedTuple('Device', [('fqdn', str), ('metadata', Mapping), ('config', Mapping), ('private', Mapping)])
logger = logging.getLogger(__name__)


class Devices(UserDict):
    """Collection of devices, accessible by FQDN as a dict or role and site via dedicated accessors."""

    def __init__(self, devices: Mapping[str, Mapping[str, str]], devices_config: Mapping[str, Mapping],
                 private_config: Optional[Mapping[str, Mapping]] = None):
        """Initialize the instance.

        Arguments:
            devices (dict): the devices configuration with FQDN as key and a dictionary with the device metadata as
                value.
            devices_config (dict): the devices configuration with FQDN as key and a dictionary with the device-specific
                configuration as value.
            private_config (dict, optional): an optional dictionary of the devices private configuration with the FQDN
                as key and a dictionary of device-specific private configuration as value. It cannot have top level
                keys in common with the same device public configuration.

        """
        super().__init__()
        if private_config is None:
            private_config = {}

        for fqdn, metadata in devices.items():
            self.data[fqdn] = Device(fqdn, metadata, devices_config.get(fqdn, {}), private_config.get(fqdn, {}))

        logger.info('Initialized %d devices', len(self.data))

    def query(self, query_string: str) -> List[Device]:
        """Get the devices matching the query.

        Todo:
            If needed, expand the query capabilities with a proper syntax using pyparsing.

        Arguments:
            query_string (str): the query_string to use to filter for.

        Raises:
            homer.exceptions.HomerError: on invalid query.

        Returns:
            list: a list of Device objects.

        """
        if ':' in query_string:  # Simple key-value query
            key, value = query_string.split(':', 1)
            results = [device for device in self.data.values() if device.metadata.get(key, None) == value]
        else:  # FQDN query
            results = [device for fqdn, device in self.items() if fnmatch.fnmatch(fqdn, query_string)]

        logger.info("Matched %d device(s) for query '%s'", len(results), query_string)
        return sorted(results, key=attrgetter('fqdn'))



"""__init__ module tests."""
import textwrap

from unittest import mock

import pytest

from jnpr.junos.exception import ConfigLoadError, RpcTimeoutError
from lxml import etree

import homer

from homer.config import load_yaml_config
from homer.tests import get_fixture_path
from homer.tests.unit.transports.test_junos import ERROR_RESPONSE


def setup_tmp_path(file_name, path):
    """Initialize the temporary directory and configuration."""
    output = path / 'output'
    output.mkdir()
    config = load_yaml_config(get_fixture_path('cli', file_name))
    config['base_paths']['output'] = str(output)
    return output, config


def get_generated_files(path):
    """Get all the generated files in the output directory."""
    return [out_file.name for out_file in path.iterdir()
            if out_file.is_file() and out_file.suffix == homer.Homer.OUT_EXTENSION]


def test_version():
    """Check that the __version__ package variable is set."""
    assert hasattr(homer, '__version__')


class TestHomer:
    """Homer class tests."""

    @pytest.fixture(autouse=True)
    def setup_method_fixture(self, tmp_path):
        """Initialize the instance."""
        # pylint: disable=attribute-defined-outside-init
        self.output, self.config = setup_tmp_path('config.yaml', tmp_path)
        self.homer = homer.Homer(self.config)

    def test_generate_ok(self):
        """It should generate the compiled configuration files for the matching devices."""
        spurious_file = self.output / f'spurious{homer.Homer.OUT_EXTENSION}'
        spurious_file.touch()  # To check that the file will be removed
        spurious_dir = self.output / 'spurious'
        spurious_dir.mkdir()

        ret = self.homer.generate('device*')

        assert ret == 0
        assert sorted(get_generated_files(self.output)) == ['device1.example.com.out', 'device2.example.com.out']
        expected = """
            roleB;
            siteB;
            device2.example.com;
            common_value;
            roleB_value;
            siteB_value;
            device2_value;
            common_private_value;
            roleB_private_value;
            siteB_private_value;
            device2_private_value;
        """
        with open(str(self.output / 'device2.example.com.out'), encoding='utf-8') as f:
            assert textwrap.dedent(expected).lstrip('\n') == f.read()

    def test_generate_no_private(self):
        """It should execute the whole program based on CLI arguments."""
        config = self.config.copy()
        del config['base_paths']['private']

        ret = homer.Homer(config).generate('device*')

        assert ret == 0
        assert sorted(get_generated_files(self.output)) == ['device1.example.com.out', 'device2.example.com.out']
        expected = """
            roleB;
            siteB;
            device2.example.com;
            common_value;
            roleB_value;
            siteB_value;
            device2_value;
        """
        with open(str(self.output / 'device2.example.com.out'), encoding='utf-8') as f:
            assert textwrap.dedent(expected).lstrip('\n') == f.read()

    def test_execute_generate_fail_to_render(self):
        """It should skip devices that fails to render the configuration."""
        ret = self.homer.generate('site:siteC')

        assert ret == 1
        assert get_generated_files(self.output) == ['valid.example.com.out']

    @pytest.mark.parametrize('diff, omit_diff, expected, ret', (
        ('', False, '# No diff', 0),
        ('', True, '# No diff', 0),
        ('some diff', False, 'some diff', 99),
        ('some diff', True, '# Non-empty diff omitted, -o/--omit-diff set', 99),
    ))
    @mock.patch('homer.transports.junos.JunOSDevice')
    def test_execute_diff_ok(  # pylint: disable=too-many-arguments
            self, mocked_device, diff, omit_diff, expected, ret, capsys):
        """It should diff the compiled configuration with the live one."""
        mocked_device.return_value.cu.diff.return_value = diff
        return_code = self.homer.diff('device*', omit_diff=omit_diff)

        out, _ = capsys.readouterr()
        assert return_code == ret
        assert mocked_device.return_value.cu.diff.called
        assert expected in out

    @mock.patch('homer.transports.junos.JunOSDevice')
    def test_execute_diff_raise(self, mocked_device, capsys, caplog):
        """It should skip the device that raises an HomerLoadError."""
        mocked_device.return_value.cu.load.side_effect = ConfigLoadError(
            etree.XML(ERROR_RESPONSE.format(insert='')))
        return_code = self.homer.diff('device1*')

        out, _ = capsys.readouterr()
        assert return_code == 1
        assert 'Failed to get diff for device1.example.com: ConfigLoadError(' in caplog.text
        assert "Changes for 1 devices: ['device1.example.com']\n# Failed" in out

    @mock.patch('builtins.input')
    @mock.patch('homer.sys.stdout.isatty')
    @mock.patch('homer.transports.junos.JunOSDevice')
    def test_execute_commit_ok(self, mocked_device, mocked_isatty, mocked_input):
        """It should commit the compiled configuration to the device."""
        # TODO: to be expanded
        mocked_isatty.return_value = True
        mocked_input.return_value = 'yes'
        ret = self.homer.commit('device*', message='commit message')
        assert ret == 0
        assert mocked_device.called

    @mock.patch('builtins.input')
    @mock.patch('homer.sys.stdout.isatty')
    @mock.patch('homer.transports.junos.JunOSDevice')
    def test_execute_commit_timeout(self, mocked_device, mocked_isatty, mocked_input, caplog):
        """It should retry TIMEOUT_ATTEMPTS times and report the failure."""
        message = 'commit message'
        mocked_device.return_value.cu.diff.return_value = 'diff'
        mocked_device.return_value.cu.commit.side_effect = RpcTimeoutError(mocked_device, message, 30)
        mocked_isatty.return_value = True
        mocked_input.return_value = 'yes'
        ret = self.homer.commit('device*', message=message)
        assert ret == 1
        assert 'Attempt 3/3 failed' in caplog.text

    @mock.patch('builtins.input')
    @mock.patch('homer.sys.stdout.isatty')
    @mock.patch('homer.transports.junos.JunOSDevice')
    @pytest.mark.parametrize('input_value, expected', (
        ('no', 'Commit aborted'),
        ('invalid', 'Too many invalid answers, commit aborted'),
    ))
    def test_execute_commit_abort(  # pylint: disable=too-many-arguments
            self, mocked_device, mocked_isatty, mocked_input, input_value, expected, caplog):
        """It should skip a device and log a warning if the commit is aborted."""
        message = 'commit message'
        mocked_isatty.return_value = True
        mocked_input.return_value = input_value
        mocked_device.return_value.cu.diff.return_value = 'diff'
        ret = self.homer.commit('device*', message=message)
        assert ret == 1
        assert expected in caplog.text
        mocked_device.return_value.cu.commit.assert_not_called()

    @mock.patch('homer.sys.stdout.isatty')
    @mock.patch('homer.transports.junos.JunOSDevice')
    def test_execute_commit_notty(self, mocked_device, mocked_isatty, caplog):
        """It should skip a device and log a warning if the commit is aborted."""
        mocked_isatty.return_value = False
        mocked_device.return_value.cu.diff.return_value = 'diff'
        ret = self.homer.commit('device*', message='commit message')
        assert ret == 1
        assert 'Not in a TTY, unable to ask for confirmation' in caplog.text
        mocked_device.return_value.cu.commit.assert_not_called()


class TestHomerNetbox:
    """Homer class tests with Netbox enabled."""

    @pytest.fixture(autouse=True)
    @mock.patch('homer.pynetbox.api')  # Pynetbox objects lazily resolve API objects, can't use autospec=True
    def setup_method_fixture(self, mocked_pynetbox, tmp_path):
        """Initialize the instance."""
        # pylint: disable=attribute-defined-outside-init
        self.output, self.config = setup_tmp_path('config-netbox.yaml', tmp_path)
        self.mocked_pynetbox = mocked_pynetbox
        self.homer = homer.Homer(self.config)

    def test_init(self):
        """The instance should have setup the Netbox API."""
        self.mocked_pynetbox.assert_called_once_with('https://netbox.example.com', token='token')  # nosec

    @mock.patch('homer.NetboxDeviceData', autospec=True)
    @mock.patch('homer.NetboxData', autospec=True)
    def test_execute_generate(self, mocked_netbox_data, mocked_netbox_device_data):
        """It should generate the configuration for the given device, including netbox data."""
        mocked_netbox_data.return_value = {'netbox_key': 'netbox_value'}
        mocked_netbox_device_data.return_value = {'netbox_key': 'netbox_device_value'}

        ret = self.homer.generate('device*')

        assert ret == 0
        assert sorted(get_generated_files(self.output)) == ['device1.example.com.out', 'device2.example.com.out']
        expected = """
            roleB;
            siteB;
            device2.example.com;
            common_value;
            roleB_value;
            siteB_value;
            device2_value;
            common_private_value;
            roleB_private_value;
            siteB_private_value;
            device2_private_value;
            netbox_value;
            netbox_device_value;
            netbox_device_plugin;
        """
        with open(str(self.output / 'device2.example.com.out'), encoding='utf-8') as f:
            assert textwrap.dedent(expected).lstrip('\n') == f.read()

    @mock.patch('homer.NetboxDeviceData', autospec=True)
    @mock.patch('homer.NetboxData', autospec=True)
    @mock.patch('homer.NetboxInventory', autospec=True)
    @mock.patch('homer.transports.junos.ConnectedDevice', autospec=True)
    @pytest.mark.parametrize('device_id, suffix, timeout', (('1', 'A', 30), ('2', 'B', 10)))
    def test_execute_diff_inventory(self, mocked_connected_device,  # pylint: disable=too-many-arguments
                                    mocked_netbox_inventory, mocked_netbox_data, mocked_netbox_device_data, device_id,
                                    suffix, timeout):
        """It should generate the configuration for the given device, including netbox data."""
        fqdn = f'device{device_id}.example.com'
        mocked_connected_device.return_value.commit_check.return_value = (True, '')
        mocked_netbox_inventory.return_value.get_devices.return_value = {
            fqdn: {
                'role': f'role{suffix}',
                'site': f'site{suffix}',
                'type': f'type{suffix}',
                'status': 'active',
                'netbox_object': mock.MagicMock(),
            }
        }
        mocked_netbox_data.return_value = {'netbox_key': 'netbox_value'}
        mocked_netbox_device_data.return_value = {'netbox_key': 'netbox_device_value'}

        ret = self.homer.diff(f'device{device_id}*')

        assert ret == 0
        mocked_connected_device.assert_called_once_with(fqdn, username='', ssh_config=None, timeout=timeout)
